
from TwistedHub import *
from FITS import *
from FITStable import *
from ReadSex import *
from masks import *
from VTKHelperFunctions import *
from basics import *
from numpy.core.numerictypes import *
import numpy
import cPickle

WFSCALE = 0.05

def rebin( d, dxa, dya, data=1, cubic=0, verb=1, constant=0, interp=1):
    if verb: print "Transforming."
    if data:
       y,x = indices(dxa.shape,float32)
       g = VTKGradMag((y+dya),1)*VTKGradMag((x+dxa),1)
    t = VTKImageTransform(d.astype(float32),dxa.astype(float32),dya.astype(float32),numret=1,interp=interp,cubic=cubic,mirror=0,constant=constant)
    if data: t=VTKMultiply(t,g)
    data = where(numpy.isnan(data),0,data)
    data = where(greater(data,-1e20),data,0)
    return t

def reversebin( d, dxa, dya, origsize=None, cubic=0, data=1, word="coordinates",verb=1,constant=0.0,interp=1):
    if verb:
       if origsize: print "Reverse Transforming to size:",origsize
       else: print "Reverse Transforming %s." % (word)
    y,x = indices(dxa.shape,d.dtype)
    if data:
       g = VTKMultiply(VTKGradMag(VTKAdd(y,dya),1),VTKGradMag(VTKAdd(x,dxa),1))
       d = VTKDivide(d,g)
    t = VTKImageTransform(d,dxa.astype(float32),dya.astype(float32),numret=1,reverse=1,origsize=origsize,interp=interp,cubic=cubic,constant=constant)
    t = where(numpy.isnan(t),0,t)
    t = where(greater(t,-1e20),t,0)
    return t

def randback(F):
   data = F.data().ravel()
   data = compress(not_equal(data,0),data)
   bg = GaussianBG(data.ravel(),299,verb=0)
   return bg

class Image:
   def __init__(self,image,subbg=0,suffix="",photoff=None,bgkey=None):
      self.image = image
      self.subbg = subbg
      self.solnfile = self.image.replace(".fits",".soln")
      self.solndat = readsoln(self.solnfile)
      self.sigim = self.image.replace(suffix+".fits","sig.fits")
      if not os.path.exists(self.sigim):
         self.sigim = self.image.replace(suffix+".fits","_sig.fits")
         if not os.path.exists(self.sigim): self.sigim = None
      self.open()
      self.naxes = self.FITS.naxes()
      self.naxis1, self.naxis2 = self.naxes
      self.exptime = self.FITS["exptime"]
      if self.exptime == "N/A": self.exptime = self.FITS["elaptime"]
      if self.exptime == "N/A": self.exptime = 1.0
      self.bg = 0.0
      if bgkey:
         self.bg = self.FITS[bgkey]
         if self.bg == "N/A": self.bg = 0.0
      if photoff:
         self.photoff = self.FITS[photoff]
         if self.photoff == "N/A": self.photoff = None
#         else: self.exptime = 1.0
      else:
         self.photoff = None
#      if self.photoff: self.exptime = 1.0
      self.ra_image,self.dec_image = self.ToSky(self.naxes[0]/2.,self.naxes[1]/2.)
      self.ras = formatSexagesimal(self.ra_image*180/pi/15,4,3600000)
      self.decs = formatSexagesimal(self.dec_image*180/pi,4,3600000)
      print "%20s %15s %15s" % (self.image,self.ras,self.decs),
      if self.photoff: print "%7.4f" % (self.photoff)
      else: print
      self.isolndat = self.invsoln(verb=1)
      if self.sigim: print "Got %s" % (self.sigim)
      self.close()

   def open(self):
      self.FITS = FITS(self.image)
      if self.sigim: self.sigFITS = FITS(self.sigim)
      else: self.sigFITS = None

   def close(self):
      self.FITS.close()
      self.FITS = None
      if self.sigim:
         self.sigFITS.close()
         self.sigFITS = None

   def GetCorners(self,cra=None,cdec=None):
       if not cra and not cdec:
          cra,cdec = self.solndat["sky"]
       self.bg = None
       self.ccors = []
       self.xcors = []
       self.ycors = []
       self.x0cors = []
       self.y0cors = []
       naxis1, naxis2 = self.naxes
       self.naxis1 = naxis1
       self.naxis2 = naxis2
       self.x0cors, self.y0cors = transpose([(xi,yi) for xi in [0,naxis1-1] for yi in [0,naxis2-1]])
       rds = [self.ToSky(xi,yi) for xi in [0,naxis1-1] for yi in [0,naxis2-1]]
       self.xcor,self.ycor = transpose([normalcoord(rds[j][0],rds[j][1],cra,cdec,conv=3600*180/pi) for j in range(4)])
       self.xcors.append(self.xcor)
       self.ycors.append(self.ycor)

   def GetBackground(self):
      self.open()
      self.bg = randback(self.FITS)
      self.close()
      print "Found background of %.1f %s." % (self.bg,self.image)

   def ToSky(self,xc,yc):
       x0,y0 = self.solndat["center"]
       rc,dc = self.solndat["sky"]
       xc = (xc-x0) + 1
       yc = (yc-y0) + 1
       u,v = mbasis(self.solndat["c"],xc,yc)
       rs, ds = equicoord(u,v,rc,dc,text=0)
       return rs,ds

   def invsoln(self,verb=0):
       nterms = len(self.solndat["c"])
       x = asarray( range(0,self.naxis1+self.naxis1/nterms,self.naxis1/nterms) + [self.naxis1])
       y = asarray( range(0,self.naxis2+self.naxis2/nterms,self.naxis2/nterms) + [self.naxis2])
       x = VTKSubtract(x, self.solndat["center"][0])
       y = VTKSubtract(y, self.solndat["center"][1])
       nx = len(x); ny = len(y)
       o = ones((ny,nx),float32)
       yy = (o*y[::,NewAxis]).ravel()
       xx = (o*x[NewAxis,::]).ravel()
       u,v = mbasis(self.solndat["c"],xx,yy)
       nterm = [4,6,12,20,30,42,56]
       nords = [0,1, 2, 4, 4, 4, 4]
       m = nords[nterm.index(nterms)]
       b = abasis(m,u,v)
       c = svdfit(b,concatenate([xx,yy]))
       xf,yf = mbasis(c,u,v)
       xres = VTKSubtract(xf,xx); yres = VTKSubtract(yf,yy)
       if verb:
         print "Using",nterm,"terms (order=",m,")"
         print "Inverse solution RMS=",rms(xres),rms(yres)
       self.solndat["ic"] = c

   def invsoln_new(self,verb=0):
       nterms = len(self.solndat["c"])
       nterm = [4,6,12,20,30,42,56]
       nords = [0,1,3,4,5,6,6]
       mo = nords[nterm.index(nterms)]
       ksp = (2*mo+2)
       x0,y0 = self.solndat["center"]
       rc,dc = self.solndat["sky"]
       imsize = self.naxis2, self.naxis1
       dyc,dxc = imsize
       ux = arange(-dxc/ksp,imsize[1]+0.5*dxc/ksp,dxc/ksp)
       uy = arange(-dyc/ksp,imsize[0]+0.5*dxc/ksp,dyc/ksp)
       x4,y4 = asarray(len(uy)*[ux]), transpose(len(ux)*[uy])
       x4,y4 = x4.ravel(), ravel(y4)
       tra,tdec = self.ToSky(x4,y4)
       tsqu,teta = normalcoord(tra,tdec,rc,dc)
       dsqu = (max(tsqu)-min(tsqu))/ksp
       deta = (max(teta)-min(teta))/ksp
       kx,ky = 3,3
       xknots = concatenate([ (kx+1)*[min(tsqu)], (kx+1)*[max(tsqu)]])
       yknots = concatenate([ (ky+1)*[min(teta)], (ky+1)*[max(teta)]])
       xb = min(xknots); xe = max(xknots); yb = min(yknots); ye = max(yknots)
       x4 = (x4 - x0) + 1
       y4 = (y4 - y0) + 1
       xtck = bisplrep(tsqu,teta,x4,task=-1,s=None,kx=kx,ky=ky,tx=xknots,ty=yknots,\
          xb=xb, xe=xe, yb=yb, ye=ye, nxest=len(xknots), nyest=len(yknots), full_output=0,quite=1)
       ytck = bisplrep(tsqu,teta,y4,task=-1,s=None,kx=kx,ky=ky,tx=xknots,ty=yknots,\
          xb=xb, xe=xe, yb=yb, ye=ye, nxest=len(xknots), nyest=len(yknots), full_output=0,quite=1)
       self.solndat["xtck"] = xtck
       self.solndat["ytck"] = ytck

   def FromSky_new(self,rs,ds):
       x0,y0 = self.solndat["center"]
       rc,dc = self.solndat["sky"]
       u, v = normalcoord(rs,ds,rc,dc)
#       xfold,yofld = self.FromSky_old(rs,ds)
       eu = extrema(self.solndat["xtck"][0])
       ev = extrema(self.solndat["xtck"][1])
       good = logical_and(between(u,eu[0],eu[1]), between(v,ev[0],ev[1]))
       xf = -999*ones(u.shape,float32)
       yf = -999*ones(v.shape,float32)
       if len(good.shape) == 1:
          if sum(good):
             u = clip(u,eu[0],eu[1])
             v = clip(v,ev[0],ev[1])
             xi,yi = gbspl(u,v,self.solndat["xtck"],self.solndat["ytck"],x0,y0)
             xu = where(good,xi,-999)
             yu = where(good,yi,-999)
             xf = xu
             yf = yu
       else:
         for j in range(len(good)):
            if sum(good[j]):
               gu = clip(u[j],eu[0],eu[1])
               gv = clip(v[j],ev[0],ev[1])
               xi,yi = gbspl(gu,gv,self.solndat["xtck"],self.solndat["ytck"],x0,y0)
               xf[j] = where(good[j],xi,-999).astype(xf.dtype)
               yf[j] = where(good[j],yi,-999).astype(yf.dtype)
       return xf, yf

   def FromSky(self,rs,ds):
       x0,y0 = self.solndat["center"]
       rc,dc = self.solndat["sky"]
       u, v = normalcoord(rs,ds,rc,dc)
       xf, yf = mbasis(self.solndat["ic"],u,v)
       xc = xf + x0 - 1
       yc = yf + y0 - 1
       return xc, yc

   def GetPixelPositions(self,ras,decs,rc=None,dc=None):
       if not rc and not dc:
          rc,dc = self.solndat["sky"]
       u, v = normalcoord(ras,decs,rc,dc,conv=3600*180/pi)
       xi, yi = self.FromSky(ras,decs)
       gx = between(xi,0,self.naxes[0],inc=1).astype(float32)
       gy = between(yi,0,self.naxes[1],inc=1).astype(float32)
       gr = between(u,min(self.xcor),max(self.xcor),inc=1)
       gd = between(v,min(self.ycor),max(self.ycor),inc=1)
       good = gx*gy*gr*gd
       return xi, yi, good.astype(float32)

   def GetSquare(self,xlo,xhi,ylo,yhi,cra,cdec,buffer=0,verb=0):
      if verb: print  "Reading %s." % (self.image),;sys.stdout.flush()
      if self.subbg:
        if self.bg == None: self.GetBackground()
      bg = self.bg
      if bg == None: bg = 0.0
      self.open()
      xlo = max([0,xlo-buffer])
      xhi = min([xhi+buffer,self.naxis1-1])
      ylo = max([0,ylo-buffer])
      yhi = min([yhi+buffer,self.naxis2-1])
      data = self.FITS.readsec([xlo,ylo],[xhi,yhi])
      data = where(numpy.isnan(data),0,data)
      data = where(between(data,-1e24,1e24),data,0.0)
      if self.sigim:
         sig = self.sigFITS.readsec([xlo,ylo],[xhi,yhi])
         sig = where(numpy.isnan(sig),0,sig)
         sig = where(between(sig,-1e24,1e24),sig,0.0)
         data = where(sig,data,0.0)
      else: sig = not_equal(data,0).astype(float32)
      if self.photoff:
        data = VTKMultiply(data,power(10,-0.4*self.photoff))
        if self.sigim: sig = VTKMultiply(sig, power(10,-0.4*self.photoff))
        bg = bg * power(10,-0.4*self.photoff)
      dq = greater(sig,0).astype(data.dtype)
      self.close()
      return (data,bg,sig,dq,xlo,ylo)

def gbspl(u,v,xtck,ytck,x0,y0):
      xc = bisplev2(u,v,xtck)
      yc = bisplev2(u,v,ytck)
      xi = xc + x0 - 1
      yi = yc + y0 - 1
      return xi,yi

def GetSquare(iy,ix,iimages,brect,srect,cra,cdec,\
     medx,medy,ucl,scl,rj,ncl,dxn,dyn,xspan,yspan,xra,ydec,scale,uniform,smooth,psf,simple,verb,interp,usetmp=0,cubic=0):
       if isinstance(iimages,ListType):
          images = iimages
       elif isinstance(iimages,StringType):
          fimages = open(iimages,"r")
          images = cPickle.load(fimages)
          fimages.close()
       ascale = scale*3600*180/pi
       au = -ascale*(1+arange(xspan)-xra)
       av = ascale*(1+arange(yspan)-ydec)
       border = 5
       y00 = max([0,iy]); y10 = min([iy + dyn,yspan])
       x00 = max([0,ix]); x10 = min([ix + dxn,xspan])
       x0 = max([0,x00-border])
       x1 = min([x10+border,xspan])
       y0 = max([0,y00-border])
       y1 = min([y10+border,yspan])
       if verb:
           print "[%d:%d,%d:%d] " % (x00,x10,y00,y10),;sys.stdout.flush()
       us1, vs1 = au[x0:x1], av[y0:y1]
       brect2,srect2 = "",""
       if not len(us1) or not len(vs1): return brect2,srect2,x00,x10,y00,y10
       us = us1[NewAxis,::] * ones(vs1.shape,float32)[::,NewAxis]
       vs = ones(us1.shape,float32)[NewAxis,::] * vs1[::,NewAxis]
       iis = range(len(images))
       nos, dos, eos, wos, ros, qos, expt, xxfm, yxfm, xxfmr, yxfmr  = [[] for kl in range(11)]
       allras,alldecs = equicoord(us*pi/180/3600,vs*pi/180/3600,cra,cdec,text=0)
       for juse in range(len(iis)):
          image = images[iis[juse]]
          cornras = asarray(extrema(allras.ravel()))
          corndecs = asarray(extrema(alldecs.ravel()))
          cornras,corndecs = transpose([[cornras[jc],corndecs[ic]] for ic in range(2) for jc in range(2)])
          cornras1,corndecs1 = transpose([[cornras[jc],corndecs[ic]-2.5*pi/180/3600] for ic in range(2) for jc in range(2)])
          cornras2,corndecs2 = transpose([[cornras[jc],corndecs[ic]+2.5*pi/180/3600] for ic in range(2) for jc in range(2)])
          corg = image.GetPixelPositions(cornras,corndecs,cra,cdec)[2]
          corg1 = image.GetPixelPositions(cornras1,corndecs1,cra,cdec)[2]
          corg2 = image.GetPixelPositions(cornras2,corndecs2,cra,cdec)[2]
#          print sum(corg) ,sum(corg1) ,sum(corg2),
          if sum(corg) or sum(corg1) or sum(corg2):
              uxi,uyi,ugi = image.GetPixelPositions(allras,alldecs,cra,cdec)
#              print sum(ugi.ravel()),
              if sum(ugi.ravel()):
                   xlo, xhi= extrema(compress(ugi.ravel(),uxi.ravel()))
                   ylo, yhi= extrema(compress(ugi.ravel(),uyi.ravel()))
                   ixlo, ixhi = int(xlo), int(xhi)
                   iylo, iyhi = int(ylo), int(yhi)
                   data,bg,sig,dq,ixlo,iylo = image.GetSquare(ixlo,ixhi,iylo,iyhi,cra,cdec,buffer=border,verb=verb)
                   if smooth > 0:
                      if verb: print "smoothing",
                      wsig = (not_equal(sig,0) * not_equal(data,0)).astype(float32)
                      denom = VTKGauss(wsig,smooth,numret=1)
                      data = divz(VTKGauss(data*wsig,smooth,numret=1),denom)
                      sig = sqrt(divz(VTKGauss(wsig*sig**2,smooth,numret=1)*smooth,denom))
                   elif smooth < 0:
                      if verb: print "smoothing",
                      wsig = (not_equal(sig,0) * not_equal(data,0)).astype(float32)
                      denom = VTKMeanSmooth(wsig,int(abs(smooth)),numret=1)
                      data = divz(VTKMeanSmooth(data,int(abs(smooth)),numret=1),denom)
                      sig = sqrt(divz(VTKMeanSmooth(sig**2,int(abs(smooth)),numret=1)*abs(smooth),denom))
                   if sum(dq.ravel()):
                      data = where(abs(data)>(1e-12*bg),VTKSubtract(data,bg),0)
                      if uniform: sig = VTKMultiply( ones(sig.shape,float32), mean(sig.ravel()))
                      oy,ox = indices(us.shape,float32)
                      xfm = (uxi-ixlo) - ox
                      yfm = (uyi-iylo) - oy
                      rawoy,rawox = indices(data.shape,float32)
                      rawrs,rawds = image.ToSky(rawox+ixlo,rawoy+iylo)
                      rawu, rawv = normalcoord(rawrs,rawds,cra,cdec,conv=3600*180/pi)
                      xfmr = (rawu-us1[0])/(us1[1]-us1[0]) - rawox
                      yfmr = (rawv-vs1[0])/(vs1[1]-vs1[0]) - rawoy
                      dos.append(data)
                      ros.append(bg)
                      eos.append(sig)
                      qos.append(dq)
                      expt.append(image.exptime)
                      nos.append(image.image)
                      xxfm.append(xfm)
                      yxfm.append(yfm)
                      xxfmr.append(xfmr)
                      yxfmr.append(yfmr)

       result = 0.0
       M = len(dos)
       if not M:
          return brect2,srect2,x00,x10,y00,y10
       else:
            if verb: print M,;sys.stdout.flush()
            for iter in range(ncl):

                if ucl and M > 1:
                   unumer = []
                   udenom = []
                   for m in range(M):
                       yco,xco = indices(xxfm[m].shape,dos[m].dtype)
                       xcfm = VTKAdd(xco,xxfm[m])
                       ycfm = VTKAdd(yco,yxfm[m])
                       gx = VTKGradMag(xcfm,1)
                       gy = VTKGradMag(ycfm,1)
                       g = VTKMultiply(gx,gy)
                       if verb: print "transforming...",;sys.stdout.flush()
                       top1 = rebin(dos[m],xxfm[m],yxfm[m],data=0,cubic=cubic,verb=0,constant=0,interp=interp)
                       err1 = rebin(VTKMultiply(eos[m],eos[m]),xxfm[m],yxfm[m],data=1,cubic=cubic,verb=0,constant=0,interp=interp)
                       top1 = VTKMultiply(top1,g)
                       top = VTKDivide(top1,expt[m])
                       bot = not_equal(err1,0)
                       unumer.append(top)
                       udenom.append(bot)

                numer, denom = [], []
                numer2, denom2 = 0.0, 0.0
#                qdump("numer.fits",unumer)
#                print expt
                for m in range(M):
                    mnumer, mdenom = [], []
                    yco,xco = indices(xxfm[m].shape,dos[m].dtype)
                    xcfm, ycfm = (xco+xxfm[m]), (yco+yxfm[m])
                    gx, gy = VTKGradMag(xcfm,1), VTKGradMag(ycfm,1)
                    g = VTKMultiply(gx,gy)
                    if ucl and M > 1:
                          top, bot = zeros(dos[m].shape,dos[m].dtype), zeros(dos[m].shape,dos[m].dtype)
                          if iter:
                             mtmpl = VTKMultiply(result,expt[m])
                             errl = None
                             if verb: print "using result",;sys.stdout.flush()
                          elif (m==0 and simple) or (not simple):
                             err2 = 0.0
                             ntemp = 0
                             for n in range(M):
                                if n != m or simple:
                                   ntemp += 1
                                   mnumer.append(unumer[n].ravel())
                                   mdenom.append(udenom[n].ravel())
                                   if verb: print "(%d,%d)" % (m,n),;sys.stdout.flush()
                             anumer = asarray(mnumer)
                             adenom = asarray(mdenom)
                             anumer = sort(where(mdenom,anumer,1e38),axis=0)
                             nnumer = add.reduce(less(anumer,1e38)).astype(int32)
                             ynumer = arange(len(mnumer))[::,NewAxis]
                             nmid = nnumer/2
                             enumer = equal(ynumer,nmid)
                             enumer2 = equal(ynumer,nmid-1)
                             neven = equal(nnumer,2*nmid)
                             nodd = logical_not(neven)
                             mtmpl = where(nodd,sum(anumer*enumer),\
                                     0.5*(sum(anumer*enumer) + sum(anumer*enumer2)))
                             mtmpl = where(less(mtmpl,1e30),mtmpl,0)
                             anumer = sort(where(mdenom,abs(anumer-mtmpl),1e38),axis=0)
                             nnumer = add.reduce(less(anumer,1e38)).astype(int32)
                             nmid = nnumer/2
                             ynumer = arange(len(mnumer))[::,NewAxis]
                             enumer = equal(ynumer,nmid)
                             enumer2 = equal(ynumer,nmid-1)
                             neven = equal(nnumer,2*nmid)
                             nodd = logical_not(neven)
                             rtmpl = where(nodd,sum(anumer*enumer),\
                                     0.5*(sum(anumer*enumer) + sum(anumer*enumer2)))
                             rtmpl = 1.48*divz(rtmpl,sqrt(nnumer))
                             mtmpl = reshape(mtmpl,unumer[0].shape)
                             rtmpl = reshape(rtmpl,unumer[0].shape)
                             rtmpl = where(less(rtmpl,1e30),rtmpl,0)
                             mnumer = reshape(mnumer,(len(mnumer),)+unumer[0].shape)
                             q = concatenate([mnumer,[mtmpl,rtmpl]],0)
#                             qdump("mnumer.fits",q)
#                             exit()
                          mtmpl_e = where(between(mtmpl,-1e24,1e24),expt[m]*mtmpl,0).astype(float32)
                          rtmpl_e = where(between(mtmpl,-1e24,1e24),expt[m]*rtmpl,0).astype(float32)
                          print "reversing...",nos[m],;sys.stdout.flush()
#                          qdump("mtmpl_e.fits",mtmpl_e)
#                          qdump("xxfmr.fits",xxfmr[m])
#                          qdump("yxfmr.fits",yxfmr[m])
#                          exit()
                          tmpl = rebin(mtmpl_e,xxfmr[m],yxfmr[m],cubic=cubic,data=1,verb=0,interp=interp,constant=0)
                          tmpl_s = rebin(rtmpl_e,xxfmr[m],yxfmr[m],cubic=cubic,data=1,verb=0,interp=interp,constant=0)
                          if medx and medy:
#                            mtmpl_m = ones(tmpl.shape)
                            medr = int(sqrt(medx*medy))
                            mtmpl_r = 1.48*median(abs(tmpl.ravel()))
                            mtmpl_m = (not_equal(tmpl,0)*greater_equal(tmpl,5*mtmpl_r)).astype(int32)
                            mtmpl_m = VTKIslandRemoval(mtmpl_m,1.0,0.0,(1+2*medr)**2,numret=1).astype(float32)
                            mtmpl_m = greater(VTKGauss(mtmpl_m,medr),0.01).astype(float32)
                            if sum(mtmpl_m.ravel()) > 0.01*len(mtmpl_m.ravel()) > 20*((2*medx+1)*(2*medy+1)):
                               print sum(mtmpl_m.ravel()),len(mtmpl_m.ravel()),;sys.stdout.flush()
                               print sum(mtmpl_m.ravel())/len(mtmpl_m.ravel()),;sys.stdout.flush()
#                               qdump("dosm.fits",dos[m])
#                               qdump("mtmpl_m.fits",mtmpl_m*dos[m])
#                               qdump("tmpl.fits",tmpl)
                               mtmpl_h = compress(mtmpl_m.ravel(),dos[m].ravel())
                               mtmpl_s = []
                               p0 = time.time()
                               for iy in range(-medy,medy+1,1):
                                 for ix in range(-medx,medx+1,1):
                                    mtmpl_b = VTKImageShift(tmpl,iy,ix,numret=1)
                                    mtmpl_s.append(compress(mtmpl_m.ravel(),mtmpl_b.ravel()))
                               if verb: print "Fitting! %s[%d:%d,%d:%d]" % (brect, x00,x10,y00,y10),;sys.stdout.flush()
                               mtmpl_f = svdfit(transpose(mtmpl_s),mtmpl_h)
                               print "Matched!"
                               mtmpl_p = reshape(mtmpl_f,(2*medy+1,2*medx+1))
                               print mtmpl_p,
                               print sum(mtmpl_p.ravel()),;sys.stdout.flush()
                               mtmpl_p = divz(mtmpl_p, sum(mtmpl_p.ravel()))
                               mtmpl_g = 0
                               rtmpl_g = 0
                               for iy in range(-medy,medy+1,1):
                                 for ix in range(-medx,medx+1,1):
                                    mtmpl_b = VTKImageShift(tmpl,iy,ix,numret=1)
                                    mtmpl_g = mtmpl_g + mtmpl_p[iy+medy,ix+medx] * mtmpl_b
                                    rtmpl_b = VTKImageShift(tmpl_s,iy,ix,numret=1)**2
                                    rtmpl_g = rtmpl_g + mtmpl_p[iy+medy,ix+medx] * rtmpl_b
                               rtmpl_g = getsqrt( rtmpl_g)
                               tmpl = mtmpl_g
                               tmpl_s = clip( rtmpl_g,tmpl_s,1e38)
                               p1 = time.time()
                               print p1-p0;sys.stdout.flush()
                          esig = sqrt(eos[m]**2 + tmpl_s**2)
                          nsig = divz(dos[m]-tmpl,sqrt(esig**2+tmpl**2))
                          zsig = 1.48*median(abs(compress(tmpl.ravel(),nsig.ravel())))
                          rsig = divz(nsig,zsig)
                          bad2 = 0.0
                          arsig = abs(rsig)
                          for isl in arange(1,scl+1,1,float32):
#                             print ucl/sqrt(isl),;sys.stdout.flush()
                             bad0 = not_equal(tmpl,0)*greater(arsig,ucl/sqrt(isl)).astype(float32)
                             bad1 = VTKIslandRemoval(bad0,1.0,0.0,int(isl),numret=1)
                             bad2 = bad2 + bad1
                          bad = greater(bad2,0).astype(float32)
                          bad = VTKConvolve(bad,k=3,numret=1).astype(float32)
                          keep = logical_or(less_equal(bad,0.25*0.25),equal(tmpl,0)).astype(dos[m].dtype)
                          if 0:
                             qdump("scratch/dosm%02d.fits" % (m+1),dos[m])
                             qdump("scratch/eosm%02d.fits" % (m+1),eos[m])
                             qdump("scratch/tmpl%02d.fits" % (m+1),tmpl)
                             qdump("scratch/nsig%02d.fits" % (m+1),nsig)
                             qdump("scratch/zsig%02d.fits" % (m+1),dos[m]*mtmpl_m)
                             qdump("scratch/rsig%02d.fits" % (m+1),rsig)
                             qdump("scratch/keep%02d.fits" % (m+1),keep*1e5)
#                          exit()
                    else:
                          keep = ones(dos[m].shape,dos[m].dtype)
                    kebm = VTKMultiply(qos[m],keep)
                    rebm = VTKDivide(VTKMultiply(dos[m],kebm),expt[m])
                    numer0 = rebin(rebm,xxfm[m],yxfm[m],data=0,cubic=cubic,verb=0,constant=0,interp=interp)
                    numer1 = VTKMultiply(numer0,g)
#                    if 1:
#                       qdump("/tmp/dos%02d.fits" % (m+1),dos[m],extras=[("object",nos[m])])
#                       qdump("/tmp/numer%02d.fits" % (m+1),numer1,extras=[("object",nos[m])])
                    denom10 = rebin(kebm,xxfm[m],yxfm[m],data=0,cubic=cubic,verb=0,constant=0,interp=interp)
                    numer1 = VTKDivide(numer1,denom10)
                    numer += [numer1]
                    wt = VTKMultiply(eos[m],kebm)
                    wt = VTKDivide(VTKMultiply(wt,wt),expt[m])
                    denom1 = rebin(wt,xxfm[m],yxfm[m],data=0,cubic=cubic,verb=0,constant=0,interp=interp)
                    denom1 = VTKMultiply(denom1,VTKMultiply(g,g))
                    denom1 = VTKDivide(denom1,denom10)
                    denom1 = VTKDivide(denom1,expt[m])
                    denom += [denom1]
                tnumer = 0.0
                tnumer2 = 0.0
                tdenom = 0.0
                tdenom2 = 0.0
                for j in range(len(numer)):
                   idenom = VTKInvert(denom[j])
                   tnumer = VTKAdd(VTKMultiply(numer[j],idenom),tnumer)
                   tdenom = VTKAdd(idenom,tdenom)
                   tnumer2 = VTKAdd(VTKMultiply(denom[j],idenom),tnumer2)
                   tdenom2 = VTKAdd(VTKNot(VTKLessEqual(idenom,0)),tdenom2)
                result = VTKDivide(tnumer,tdenom)
                if srect:
                   sigres = VTKSqrt(VTKDivide(VTKDivide(tnumer2,tdenom),tdenom2))
#                   sigres = VTKMultiply(VTKSqrt(VTKDivide(VTKDivide(tnumer2,tdenom),tdenom2)),1/sqrt(2.))

            if sum(result.ravel()) == 0.0:
               print "zero!"
               brect2,srect2 = "",""
               return brect2,srect2,x00,x10,y00,y10
            else:
               result = 1*result[y00-y0:result.shape[0]+y10-y1,x00-x0:result.shape[1]+x10-x1]
               if sum(result.ravel()) == 0.0:
                   print "zero!"
                   brect2,srect2 = "",""
                   return brect2,srect2,x00,x10,y00,y10
               else:
                   if verb: print "writing %s[%d:%d,%d:%d]" % (brect, x00,x10,y00,y10),;sys.stdout.flush()
                   brect2,srect2 = "",""
                   if usetmp:
                      brect2 = os.path.dirname(brect)+"/."+".".join(os.path.basename(brect).split(".")[:-1])+"_%06d_%06d.fits" % (x00,y00)
                      qdump(brect2,result,verb=0)
                      if srect:
                         sigres = 1*sigres[y00-y0:sigres.shape[0]+y10-y1,x00-x0:sigres.shape[1]+x10-x1]
                         srect2 = os.path.dirname(srect)+"/."+".".join(os.path.basename(srect).split(".")[:-1])+"_%06d_%06d.fits" % (x00,y00)
                         qdump(srect2,sigres,verb=0)
                   else:
                      newout = FITS(brect,readonly=0)
                      newout.writesec(result,[x00,y00],[x10-1,y10-1])
                      newout.flush()
                      newout.close()
                      if srect:
                         sigres = 1*sigres[y00-y0:sigres.shape[0]+y10-y1,x00-x0:sigres.shape[1]+x10-x1]
                         sigout = FITS(srect,readonly=0)
                         sigout.writesec(sigres,[x00,y00],[x10-1,y10-1])
                         sigout.flush()
                         sigout.close()
                   return brect2,srect2,x00,x10,y00,y10

def PasteIn(brect,srect,x00,y00,clean=0,st=0,PRE=None,POST=None):
    brect2,srect2,x00,x10,y00,y10 = PRE
    if brect2:
#       print brect2
       PasteInOne(brect,brect2,x00,y00,clean=clean,st=0.0,tw=0)
       PasteInOne(srect,srect2,x00,y00,clean=clean,st=0.0,tw=0)

def PasteInOne(rect,rect2,x00,y00,clean=0,st=0,tw=0,verb=0,dnext=None):
    if rect and rect2:
#       rect2 = os.path.dirname(rect)+"/."+".".join(os.path.basename(rect).split(".")[:-1])+"_%06d_%06d.fits" % (x00,y00)
       if st == 0.0: st = time.time()
       dt = time.time()-st
#       if not FileCheck(rect2):
#          print rect2,dt,tw
#          if dt < 10.0:
#             reactor.callLater(0.1,PasteInOne,rect,x00,y00,clean=clean,st=st,tw=tw+1,dnext=dnext)
#          else:
#             print "This was missing (%s)" %(rect)
#             print rect2,FileCheck(rect2),dt,tw
#             if dnext != None: reactor.callLater(0.1,dnext.callback,True)
#       else:
       if 1:
          F = FITS(rect2)
          d = F.data()
          F.close()
          x10 = x00 + d.shape[1]
          y10 = y00 + d.shape[0]
          F = FITS(rect,readonly=0,locking=False)
          F.writesec(d,[x00,y00],[x10-1,y10-1])
          F.close()
          if clean: os.remove(rect2)
#          if dnext != None: reactor.callLater(0.1,dnext.callback,True)
#    else:
#        if dnext != None: reactor.callLater(0.1,dnext.callback,True)


