
from basics import *
from numpy.core.numerictypes import *
import numpy
from numexpr import *

def L2(x,w=0):
    if w:
       return ones(x.shape,x.dtype)
    else:
       if isinstance(x,ArrayType):
          return evaluate("0.5*(x**2)")
       else:
          return 0.5*power(x,2)

def L1(x,w=0):
    if w:
       return divz(1,abs(x))
    else:
       return abs(x)

def L1L2(x,w=0):
    if w:
       if isinstance(x,ArrayType):
          return evaluate("1/sqrt(1+x*x/2)")
       else:
          return 1/sqrt(1+x*x/2)
    else:
       if isinstance(x,ArrayType):
          return evaluate("2*(sqrt(1+x*x/2)-1)")
       else:
          return 2*(sqrt(1+x*x/2)-1)

def Lp(x,p=1.2,w=0):
    if w:
       b = equal(x,0)
       if isinstance(x,ArrayType):
          return evaluate("(b+abs(x))**(p-2)*(1-b)")
       else:
          return power(b+abs(x),p-2)*(1-b)
    else:
       if isinstance(x,ArrayType):
          return evaluate("(abs(x)**p)/p")
       else:
          return power(abs(x),p)/p

def Fair(x,c=1.3998,w=0):
    if w:
       if isinstance(x,ArrayType):
          return evaluate("1/(1+abs(x)/c)")
       else:
          return 1/(1+abs(x)/c)
    else:
       if isinstance(x,ArrayType):
          return evaluate("c*c*(abs(x)/c - log(1+abs(x)/c))")
       else:
          return c*c*(abs(x)/c - log(1+abs(x)/c))

def Cauchy(x,c=2.3849,w=0):
    if w:
       if isinstance(x,ArrayType):
          return evaluate("1/(1+(x/c)**2)")
       else:
          return 1/(1+(x/c)**2)
    else:
       if isinstance(x,ArrayType):
          return evaluate("(c**2/2)*log(1+(x/c)**2)")
       else:
          return (c**2/2)*log(1+(x/c)**2)

def Huber(x,c=1.345,w=0):
    if w:
       if isinstance(x,ArrayType):
          h0 = evaluate("(abs(x)<c)")
          h1 = ones(x.shape,x.dtype)
          h2 = evaluate("c/(abs(x)+(x==0))*(x!=0)")
          return evaluate("where(h0,h1,h2)")
       else:
          h0 = less(abs(x),c)
          h1 = ones(x.shape,x.dtype)
          h2 = divz(c,abs(x))
          return where(h0,h1,h2)
    else:
       if isinstance(x,ArrayType):
          h0 = evaluate("(abs(x)<c)")
          h1 = evaluate("0.5*(x**2)")
          h2 = evaluate("c*(abs(x)-c/2.)")
          return evaluate("where(h0,h1,h2)")
       else:
          h0 = less(abs(x),c)
          h1 = 0.5*power(x,2)
          h2 = c*(abs(x)-c/2.)
          return where(h0,h1,h2)

def Huber2(x,c=1.2107,w=0):
    if isinstance(x,ArrayType):
       h0 = evaluate("(abs(x)<c*pi/2.)")
    else:
       h0 = less(abs(x),c*pi/2.)
    if w:
       if isinstance(x,ArrayType):
          h1b = evaluate("x/c")
          h1 = evaluate("sin(h1b)/(h1b+(h1b==0))*(h1b!=0)+1.0*(h1b==0)")
          h2 = evaluate("c/(abs(x)+(x==0))*(x!=0)")
          return evaluate("where(h0,h1,h2)")
       else:
          h1 = divz(sin(x/c),x/c,repl=1)
          h2 = divz(c,abs(x))
          return where(h0,h1,h2)
    else:
       if isinstance(x,ArrayType):
          h1 = evaluate("c**2*(1-cos(x/c))")
          h2 = evaluate("c*abs(x) + c**2*(1-pi/2)")
          return evaluate("where(h0,h1,h2)")
       else:
          h1 = c**2*(1-cos(x/c))
          h2 = c*abs(x) + c**2*(1-pi/2)
          return where(h0,h1,h2)

def Welsch(x,c=2,w=0):
    if w:
       return exp(-(x/c)**2)
    else:
       return (c**2/2)*(1-exp(-(x/c)**2))

def Tukey(x,c=4.6851,w=0):
    h0 = less(abs(x),c)
    if w:
       h1 = (1-power(x/c,2))**2
       h2 = zeros(x.shape,float64)
    else:
       h1 = c**2/6. * (1-(1-power(x/c,2))**3)
       h2 = c**2/6.
    return where(h0,h1,h2)

